from typing import Union, Callable
from collections.abc import Iterable

import numpy as np
import scipy.sparse as sp
import sparse

from libertem.utils import make_polar

# Import here for backwards compatibility, refs #1031
from libertem.common.sparse import to_dense, to_sparse, is_sparse  # NOQA: F401

MaskArrayType = Union[np.ndarray, sp.coo_matrix, sp.dok_matrix]
MaskFactoriesType = Union[Callable[[], MaskArrayType], Iterable[Callable[[], MaskArrayType]]]


def _make_circular_mask(centerX, centerY, imageSizeX, imageSizeY, radius, antialiased=False):
    """
    Make a circular mask in a bool array for masking a region in an image.

    Parameters
    ----------
    centerX, centerY : float
        Centre point of the mask.
    imageSizeX, imageSizeY : int
        Size of the image to be masked.
    radius : float
        Radius of the mask.

    Returns
    -------
    Boolean Numpy 2D Array
        Array with the shape (imageSizeX, imageSizeY) with the mask.

    Examples
    --------

    >>> image = np.ones((9, 9))
    >>> mask = _make_circular_mask(4, 4, 9, 9, 2)
    >>> image_masked = image*mask
    >>> import matplotlib.pyplot as plt
    >>> cax = plt.imshow(image_masked)
    """
    if antialiased:
        mask = radial_bins(
            centerX, centerY, imageSizeX, imageSizeY, radius, n_bins=1, use_sparse=False
        )[0]
    else:
        x, y = np.ogrid[-centerY:imageSizeY-centerY, -centerX:imageSizeX-centerX]
        mask = x*x + y*y <= radius*radius
    return mask


def sparse_template_multi_stack(mask_index, offsetX, offsetY, template, imageSizeX, imageSizeY):
    '''
    Stamp the template in a multi-mask 3D stack at the positions indicated by
    mask_index, offsetY, offsetX. The function clips the bounding box as necessary.
    '''
    num_templates = len(mask_index)
    fy, fx = template.shape
    area = fy * fx
    total_index_size = num_templates * area
    y, x = np.mgrid[0:fy, 0:fx]

    data = np.zeros(total_index_size, dtype=template.dtype)
    coord_mask = np.zeros(total_index_size, dtype=int)
    coord_y = np.zeros(total_index_size, dtype=int)
    coord_x = np.zeros(total_index_size, dtype=int)

    for i in range(len(mask_index)):
        start = i * area
        stop = (i + 1) * area
        data[start:stop] = template.flatten()
        coord_mask[start:stop] = mask_index[i]
        coord_y[start:stop] = y.flatten() + offsetY[i]
        coord_x[start:stop] = x.flatten() + offsetX[i]

    selector = (coord_y >= 0) * (coord_y < imageSizeY) * (coord_x >= 0) * (coord_x < imageSizeX)

    return sparse.COO(
        data=data[selector],
        coords=(coord_mask[selector], coord_y[selector], coord_x[selector]),
        shape=(int(max(mask_index) + 1), imageSizeY, imageSizeX)
    )


def sparse_circular_multi_stack(mask_index, centerX, centerY, imageSizeX, imageSizeY, radius):
    # we make sure it is odd
    bbox = int(2*np.ceil(radius) + 1)
    bbox_center = int((bbox - 1) // 2)
    template = circular(
        centerX=bbox_center,
        centerY=bbox_center,
        imageSizeX=bbox,
        imageSizeY=bbox,
        radius=radius)
    return sparse_template_multi_stack(
        mask_index=mask_index,
        offsetX=np.array(centerX, dtype=int) - bbox_center,
        offsetY=np.array(centerY, dtype=int) - bbox_center,
        template=template,
        imageSizeX=imageSizeX,
        imageSizeY=imageSizeY,
    )


def circular(centerX, centerY, imageSizeX, imageSizeY, radius, antialiased=False):
    """
    Make a circular mask as a 2D array

    Parameters
    ----------
    centreX, centreY : float
        Centre point of the mask.
    imageSizeX, imageSizeY : int
        Size of the image to be masked.
    radius : float
        Radius of the mask.

    Returns
    -------
    Numpy 2D Array
        Array with the shape (imageSizeX, imageSizeY) with the mask.
    """
    mask = _make_circular_mask(centerX, centerY, imageSizeX, imageSizeY, radius, antialiased)
    return mask


def ring(centerX, centerY, imageSizeX, imageSizeY, radius, radius_inner, antialiased=False):
    """
    Make a ring mask as a double array.

    Parameters
    ----------
    centreX, centreY : float
        Centre point of the mask.
    imageSizeX, imageSizeY : int
        Size of the image to be masked.
    radius : float
        Outer radius of the ring.
    radius_inner : float
        Inner radius of the ring.

    Returns
    -------
    Numpy 2D Array
        Array with the shape (imageSizeX, imageSizeY) with the mask.
    """
    if antialiased:
        mask = radial_bins(
            centerX, centerY, imageSizeX, imageSizeY,
            radius=radius, radius_inner=radius_inner, n_bins=1, use_sparse=False
        )[0]
    else:
        outer = _make_circular_mask(centerX, centerY, imageSizeX, imageSizeY, radius)
        inner = _make_circular_mask(centerX, centerY, imageSizeX, imageSizeY, radius_inner)
        mask = outer & ~inner
    return mask


def radial_gradient(centerX, centerY, imageSizeX, imageSizeY, radius, antialiased=False):
    '''
    Generate a linear radial gradient from 0 to 1 within radius
    '''
    x, y = np.ogrid[-centerY:imageSizeY-centerY, -centerX:imageSizeX-centerX]
    if antialiased:
        r = np.sqrt(x**2 + y**2)
        mask = radial_gradient_background_subtraction(
            r=r, r0=radius, r_outer=0
        )
    else:
        mask = (x*x + y*y <= radius*radius) * (np.sqrt(x*x + y*y) / radius)
    return mask


def radial_gradient_background_subtraction(r, r0, r_outer, delta=1):
    '''
    Generate a template with a linear radial gradient from 0 to 1 inside r0,
    linear transition region for antialiasing between [r0 - delta/2, r0 + delta/2[,
    and a negative ring with value -1 in [r0 + delta/2, r_outer].

    The function accepts the radius for each pixel as a parameter so that a distorted version can
    be generated with the stretchY and angle parameters of :meth:`~libertem.masks.polar_map`.

    Parameters
    ----------

    r : numpy.ndarray
        Map of radius for each pixel, typically 2D. This allows to work in distorted coordinate
        systems by assigning arbitrary radius values to each pixel.
        :meth:`~libertem.masks.polar_map` can generate elliptical maps as an example.
    r0 : float
        Inner radius to fill with a linear gradient in units of r
    r_outer : float
        Outer radius of ring from r0 to fill with -1 in units of r
    delta : float, optional
        Width of transition region between inner and outer in units of r
        with linear gradient for antialiasing or smoothening. Defaults to 1.

    Returns
    -------

    numpy.ndarray
        NumPy numpy.ndarray with the same shape and type of r with mask values assigned as
        described in the description.

    '''
    result = np.zeros_like(r)
    within = r < r0 - delta/2
    result[within] = r[within] / r0

    transition = (r >= r0 - delta/2) * (r < r0 + delta/2)
    result[transition] = (r0 - r[transition]) / (delta/2)

    without = (r >= r0 + delta/2) * (r <= r_outer)
    result[without] = -1

    return result


def polar_map(centerX, centerY, imageSizeX, imageSizeY, stretchY=1., angle=0.):
    '''
    Return a map of radius and angle.

    The optional parameters stretchY and angle allow to stretch and rotate the coordinate system
    into an elliptical form. This is useful to generate modified input data for functions that
    generate a template as a function of radius and angle.

    Parameters
    ----------

    centerX,centerY : float
        Center of the coordinate system in pixel coordinates
    imageSizeX,imageSizeY : int
        Size of the map to generate in px
    stretchY,angle : float, optional
        Stretch the radius elliptically by amount :code:`stretchY` in direction
        :code:`angle` in radians. :code:`angle = 0` means in Y direction.

    Returns
    -------

    Tuple[numpy.ndarray, numpy.ndarray]
        Map of radius and angle of shape :code:`(imageSizeY, imageSizeX)`
    '''
    y, x = np.mgrid[0:imageSizeY, 0:imageSizeX]
    dy = y - centerY
    dx = x - centerX
    if stretchY != 1.0 or angle != 0.:
        (dy, dx) = (
            (dy*np.cos(angle) - dx*np.sin(angle)) / stretchY,
            dx*np.cos(angle) + dy*np.sin(angle),
        )

    dy = dy.flatten()
    dx = dx.flatten()
    cartesians = np.stack((dy, dx)).T
    polars = make_polar(cartesians)
    return (
        polars[:, 0].reshape((imageSizeY, imageSizeX)),
        polars[:, 1].reshape((imageSizeY, imageSizeX))
    )


def balance(template):
    '''
    Accept a template with both positive and negative values and scale the negative
    part in such a way that the sum is zero.

    This is useful to generate masks that return zero when applied to a
    uniform background or linear gradient.
    '''
    result = template.copy()
    above = template > 0
    below = template < 0
    result[below] *= template[above].sum() / template[below].sum() * -1
    return result


def bounding_radius(centerX, centerY, imageSizeX, imageSizeY):
    '''
    Calculate a radius around centerX, centerY that covers the whole frame
    '''
    dy = max(centerY, imageSizeY - centerY)
    dx = max(centerX, imageSizeX - centerX)
    return int(np.ceil(np.sqrt(dy**2 + dx**2))) + 1


def radial_bins(centerX, centerY, imageSizeX, imageSizeY,
        radius=None, radius_inner=0, n_bins=None, normalize=False, use_sparse=None, dtype=None):
    '''
    Generate antialiased rings
    '''
    if radius is None:
        radius = bounding_radius(centerX, centerY, imageSizeX, imageSizeY)

    if n_bins is None:
        n_bins = int(np.round(radius - radius_inner))

    r, phi = polar_map(centerX, centerY, imageSizeX, imageSizeY)
    r = r.flatten()

    width = (radius - radius_inner) / n_bins
    bin_area = np.pi * (radius**2 - (radius - width)**2)

    if use_sparse is None:
        use_sparse = bin_area / (imageSizeX * imageSizeY) < 0.1

    if use_sparse:
        jjs = np.arange(len(r), dtype=np.int64)

    slices = []
    for r0 in np.linspace(radius_inner, radius - width, n_bins) + width/2:
        diff = np.abs(r - r0)
        # The "0.5" ensures that the bins overlap and sum up to exactly 1
        vals = np.maximum(0, np.minimum(1, width/2 + 0.5 - diff))
        if use_sparse:
            select = vals != 0
            vals = vals[select]
            if normalize:  # Make sure each bin has a sum of 1
                s = vals.sum()
                if not np.isclose(s, 0):
                    vals /= s
            slices.append(sparse.COO(shape=len(r), data=vals.astype(dtype), coords=(jjs[select],)))
        else:
            if normalize:  # Make sure each bin has a sum of 1
                s = vals.sum()
                if not np.isclose(s, 0):
                    vals /= s
            slices.append(vals.reshape((imageSizeY, imageSizeX)).astype(dtype))
    # Patch a singularity at the center
    if radius_inner < 0.5:
        yy = int(np.round(centerY))
        xx = int(np.round(centerX))
        if yy >= 0 and yy < imageSizeY and xx >= 0 and xx < imageSizeX:
            if use_sparse:
                index = yy * imageSizeX + xx
                diff = 1 - slices[0][index] - radius_inner
                patch = sparse.COO(shape=len(r), data=[diff], coords=[index])
                slices[0] += patch
            else:
                slices[0][yy, xx] = 1 - radius_inner
    if use_sparse:
        return sparse.stack(slices).reshape((-1, imageSizeY, imageSizeX))
    else:
        return np.stack(slices)


def background_subtraction(centerX, centerY, imageSizeX, imageSizeY, radius, radius_inner,
        antialiased=False):
    mask_1 = circular(
        centerX, centerY, imageSizeX, imageSizeY, radius_inner, antialiased=antialiased
    )
    sum_1 = np.sum(mask_1)
    mask_2 = ring(
        centerX, centerY, imageSizeX, imageSizeY, radius, radius_inner, antialiased=antialiased
    )
    sum_2 = np.sum(mask_2)
    mask = mask_1 - mask_2*sum_1/sum_2
    return mask


def rectangular(X, Y, Width, Height, imageSizeX, imageSizeY):
    """
    Make a rectangular mask as a 2D array of bool.
    Parameters
    ----------
    X, Y : Corner coordinates
        Centre point of the mask.
    imageSizeX, imageSizeY : int
        Size of the image to be masked.
    Width, Height : Width and Height of the rectangle
    Returns
    -------
    Numpy 2D Array
        Array with the shape (imageSizeX, imageSizeY) with the mask.
    """
    bool_mask = np.zeros([imageSizeY, imageSizeX], dtype="bool")
    if Height*Width > 0:
        ymin = min(Y, Y+Height)
        xmin = min(X, X+Width)
        ymax = max(Y, Y+Height)
        xmax = max(X, X+Width)
    elif Height > 0 and Width < 0:
        ymin = Y
        xmin = X+Width
        ymax = Y+Height
        xmax = X
    elif Height < 0 and Width > 0:
        ymin = Y+Height
        xmin = X
        ymax = Y
        xmax = X+Width
    else:
        ymin = 0
        xmin = 0
        ymax = -1
        xmax = -1
    ymin = int(ymin)
    xmin = int(xmin)
    ymax = int(ymax)
    xmax = int(xmax)
    bool_mask[max(0, ymin):min(ymax+1, imageSizeY), max(0, xmin):min(xmax+1, imageSizeX)] = 1
    return bool_mask


# TODO: dtype parameter? consistency with ring/circular above
def gradient_x(imageSizeX, imageSizeY, dtype=np.float32):
    return np.tile(
        np.ogrid[slice(0, imageSizeX)].astype(dtype), imageSizeY
    ).reshape(imageSizeY, imageSizeX)


def gradient_y(imageSizeX, imageSizeY, dtype=np.float32):
    return gradient_x(imageSizeY, imageSizeX, dtype).transpose()



import os

from scipy.sparse import csr_matrix
import numpy as np
import pytest
from numpy.testing import assert_allclose

from libertem.io.dataset import raw_csr
from libertem.io.dataset.raw_csr import (
    read_tiles_straight, read_tiles_with_roi, CSRTriple, RawCSRDataSet
)
from sparseconverter import NUMPY, for_backend, SCIPY_CSR
from libertem.io.dataset.base import TilingScheme
from libertem.common import Shape, Slice
from libertem.udf.sum import SumUDF
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.masks import ApplyMasksUDF
from libertem.udf.stddev import StdDevUDF
from libertem.common.math import prod
from libertem.web.dataset import prime_numba_cache

from utils import _mk_random, get_testdata_path

RAW_CSR_TESTDATA_PATH = os.path.join(get_testdata_path(), 'raw_csr', 'sparse.toml')
HAVE_CSR_TESTDATA = os.path.exists(RAW_CSR_TESTDATA_PATH)


def test_get_tiles_straight():
    data = _mk_random((13, 17, 24, 19), array_backend=NUMPY)
    data_flat = data.reshape((13*17, 24*19))

    sig_shape = (24, 19)
    tileshape = Shape((11, ) + sig_shape, sig_dims=2)
    dataset_shape = Shape(data.shape, sig_dims=2)

    orig = for_backend(data_flat, SCIPY_CSR)

    orig_triple = CSRTriple(indptr=orig.indptr, indices=orig.indices, data=orig.data)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset_shape,
        intent='tile'
    )
    partition_slice = Slice(origin=(3, 0, 0), shape=Shape((51, ) + sig_shape, sig_dims=2))

    ref = for_backend(np.sum(data_flat[3:3+51], axis=0), NUMPY)

    res = np.zeros((1, data_flat.shape[1]), dtype=data.dtype)

    for tile in read_tiles_straight(
            triple=orig_triple,
            partition_slice=partition_slice,
            tiling_scheme=tiling_scheme,
            dest_dtype=res.dtype,
            sync_offset=0
    ):
        res += np.sum(tile.data, axis=0)
        assert tile.shape[0] <= tuple(tileshape)[0]
        assert isinstance(tile.data, csr_matrix)

    assert np.allclose(ref, res)


def test_get_tiles_simple():
    data = np.array((
        [1, 1, 1],
        [2, 2, 2],
        [3, 3, 3],
        [4, 4, 4],
    ))
    data_flat = data.reshape((4, 3))

    sig_shape = (3, )
    tileshape = Shape((1, ) + sig_shape, sig_dims=1)
    dataset_shape = Shape(data.shape, sig_dims=1)

    orig = for_backend(data_flat, SCIPY_CSR)

    orig_triple = CSRTriple(indptr=orig.indptr, indices=orig.indices, data=orig.data)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset_shape,
        intent='tile'
    )
    partition_slice = Slice(origin=(1, 0), shape=Shape((3, ) + sig_shape, sig_dims=1))

    ref = for_backend(np.sum(data_flat[1:], axis=0), NUMPY)

    res = np.zeros((1, data_flat.shape[1]), dtype=data.dtype)

    for tile in read_tiles_straight(
            triple=orig_triple,
            partition_slice=partition_slice,
            tiling_scheme=tiling_scheme,
            dest_dtype=res.dtype,
            sync_offset=0
    ):
        res += np.sum(tile.data, axis=0)
        assert tile.shape[0] <= tuple(tileshape)[0]
        assert isinstance(tile.data, csr_matrix)

    assert np.allclose(ref, res)


@pytest.mark.with_numba
def test_get_tiles_simple_roi():
    data = np.array((
        [1, 1, 1],
        [2, 2, 2],
        [3, 3, 3],
        [4, 4, 4],
    ))
    data_flat = data.reshape((4, 3))

    sig_shape = (3, )
    tileshape = Shape((1, ) + sig_shape, sig_dims=1)
    dataset_shape = Shape(data.shape, sig_dims=1)

    orig = for_backend(data_flat, SCIPY_CSR)

    orig_triple = CSRTriple(indptr=orig.indptr, indices=orig.indices, data=orig.data)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset_shape,
        intent='tile'
    )
    partition_slice = Slice(origin=(1, 0), shape=Shape((3, ) + sig_shape, sig_dims=1))
    roi = np.ones(data.shape[0], dtype=bool)
    roi[0:2] = 0
    roi[-1] = 0

    ref = for_backend(np.sum(data_flat[2:-1], axis=0), NUMPY)

    res = np.zeros((1, data_flat.shape[1]), dtype=data.dtype)

    for tile in read_tiles_with_roi(
            triple=orig_triple,
            partition_slice=partition_slice,
            tiling_scheme=tiling_scheme,
            roi=roi,
            dest_dtype=res.dtype,
            sync_offset=0
    ):
        res += np.sum(tile.data, axis=0)
        assert tile.shape[0] <= tuple(tileshape)[0]
        assert isinstance(tile.data, csr_matrix)

    assert np.allclose(ref, res)


def test_get_tiles_roi():
    data = _mk_random((13, 17, 24, 19), array_backend=NUMPY)
    data_flat = data.reshape((13*17, 24*19))

    sig_shape = (24, 19)
    tileshape = Shape((11, ) + sig_shape, sig_dims=2)
    dataset_shape = Shape(data.shape, sig_dims=2)

    orig = for_backend(data_flat, SCIPY_CSR)

    orig_triple = CSRTriple(indptr=orig.indptr, indices=orig.indices, data=orig.data)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset_shape,
        intent='tile'
    )
    roi = np.random.choice([True, False], data_flat.shape[0])
    partition_slice = Slice(origin=(3, 0, 0), shape=Shape((51, ) + sig_shape, sig_dims=2))
    roi_slice = partition_slice.get(roi, nav_only=True)

    ref = for_backend(np.sum(data_flat[3:3+51][roi_slice], axis=0), NUMPY)

    res = np.zeros((1, data_flat.shape[1]), dtype=data.dtype)
    total_size = 0
    for tile in read_tiles_with_roi(
            triple=orig_triple,
            partition_slice=partition_slice,
            tiling_scheme=tiling_scheme,
            roi=roi,
            dest_dtype=res.dtype,
            sync_offset=0
    ):
        res += np.sum(tile.data, axis=0)

        total_size += tile.data.shape[0]
        assert tile.shape[0] <= tuple(tileshape)[0]
        assert isinstance(tile.data, csr_matrix)
    assert total_size == np.count_nonzero(roi_slice)

    assert np.allclose(ref, res)


@pytest.mark.parametrize(
    'endian', ('native', 'big')
)
def test_raw_csr_ds_sum(
        endian, raw_csr_generated, mock_sparse_data,
        raw_csr_generated_bigendian, lt_ctx):
    if endian == 'native':
        _, data_flat = mock_sparse_data
        ds = raw_csr_generated
    elif endian == 'big':
        _, data_flat = mock_sparse_data
        ds = raw_csr_generated_bigendian
        assert ds.dtype == np.dtype('>i4')
    else:
        raise ValueError()
    udf = SumUDF()
    res = lt_ctx.run_udf(udf=udf, dataset=ds)
    ref = for_backend(np.sum(data_flat, axis=0), NUMPY)
    assert np.allclose(ref, res['intensity'].data.reshape((-1,)))


@pytest.mark.parametrize(
    'endian', ('native', 'big')
)
def test_raw_csr_ds_sum_roi(
        endian, raw_csr_generated, mock_sparse_data,
        raw_csr_generated_bigendian, lt_ctx):
    if endian == 'native':
        orig, data_flat = mock_sparse_data
        ds = raw_csr_generated
    elif endian == 'big':
        orig, data_flat = mock_sparse_data
        ds = raw_csr_generated_bigendian
        assert ds.dtype == np.dtype('>i4')
    else:
        raise ValueError()
    udf = SumUDF()
    roi = np.random.choice([True, False], data_flat.shape[0])
    res = lt_ctx.run_udf(udf=udf, dataset=ds, roi=roi)
    ref = for_backend(np.sum(data_flat[roi], axis=0), NUMPY)
    assert np.allclose(ref, res['intensity'].data.reshape((-1,)))


def test_raw_csr_ds_sumsig(raw_csr_generated, mock_sparse_data, lt_ctx):
    orig, data_flat = mock_sparse_data
    udf = SumSigUDF()
    res = lt_ctx.run_udf(udf=udf, dataset=raw_csr_generated)
    ref = for_backend(np.sum(data_flat, axis=(-1,)), NUMPY)
    assert np.allclose(ref, res['intensity'].data.reshape((-1,)))


def test_raw_csr_ds_sumsig_roi(raw_csr_generated, mock_sparse_data, lt_ctx):
    _, data_flat = mock_sparse_data
    udf = SumSigUDF()
    roi = np.random.choice([True, False], data_flat.shape[0])
    print("ROI size", np.count_nonzero(roi))
    # import pdb; pdb.set_trace()
    res = lt_ctx.run_udf(udf=udf, dataset=raw_csr_generated, roi=roi)
    ref = for_backend(np.sum(data_flat[roi], axis=(-1,)), NUMPY)
    assert np.allclose(ref, res['intensity'].raw_data.reshape((-1,)))


@pytest.fixture(scope="function")
def real_csr_data(lt_ctx):
    yield lt_ctx.load("raw_csr", path=RAW_CSR_TESTDATA_PATH)


@pytest.mark.slow
@pytest.mark.skipif(not HAVE_CSR_TESTDATA, reason="need raw CSR testdata")
def test_sum_real_data(real_csr_data, local_cluster_ctx):
    udf = SumUDF()

    local_cluster_ctx.run_udf(udf=udf, dataset=real_csr_data)
    # ref = for_backend(np.sum(data_flat[roi], axis=0), NUMPY)
    # assert np.allclose(ref, res['intensity'].data.reshape((-1,)))


@pytest.mark.slow
@pytest.mark.skipif(not HAVE_CSR_TESTDATA, reason="need raw CSR testdata")
def test_sum_real_data_roi(real_csr_data, local_cluster_ctx):
    udf = SumUDF()

    roi = np.random.choice([True, False], real_csr_data.shape.nav)
    local_cluster_ctx.run_udf(udf=udf, dataset=real_csr_data, roi=roi)
    # ref = for_backend(np.sum(data_flat[roi], axis=0), NUMPY)
    # assert np.allclose(ref, res['intensity'].data.reshape((-1,)))


def test_detect_params(raw_csr_generated, default_raw, inline_executor):
    assert RawCSRDataSet.detect_params(raw_csr_generated._path, inline_executor)
    assert not RawCSRDataSet.detect_params(default_raw._path, inline_executor)


def test_diagnostics(raw_csr_generated):
    diags = raw_csr_generated.diagnostics
    assert {"name": "data dtype", "value": "float32"} in diags
    assert {"name": "indptr dtype", "value": "int32"} in diags
    assert {"name": "indices dtype", "value": "int32"} in diags


def test_exception_at_detect(tmpdir_factory, dask_executor):
    # setup: detect on valid UFT8, invalid TOML
    datadir = tmpdir_factory.mktemp('raw_csr_txt')
    fn = str(datadir / 'test.txt')
    with open(fn, "w") as f:
        f.write("stuff,in,here")

    # exceptions should be properly caught and should be pickleable:
    assert RawCSRDataSet.detect_params(fn, executor=dask_executor) is False


def test_sig_nav_shape(raw_csr_generated, lt_ctx):
    assert len(tuple(raw_csr_generated.shape.nav)) > 1
    assert len(tuple(raw_csr_generated.shape.sig)) > 1

    flat_nav = [prod(tuple(raw_csr_generated.shape.nav))]
    # throw in dict keys iterable for variety to test robustness for
    # heterogeneous types
    flat_sig = {
        prod(tuple(raw_csr_generated.shape.sig)): None
    }.keys()

    # Confirm types are incompatible
    with pytest.raises(TypeError):
        flat_nav + flat_sig

    ds = lt_ctx.load('auto', path=raw_csr_generated._path, nav_shape=flat_nav, sig_shape=flat_sig)
    udf = SumSigUDF()
    res = lt_ctx.run_udf(dataset=ds, udf=udf)
    ref = lt_ctx.run_udf(dataset=raw_csr_generated, udf=udf)

    assert np.allclose(
        ref['intensity'].data,
        res['intensity'].data.reshape(raw_csr_generated.shape.nav)
    )


@pytest.mark.slow
@pytest.mark.parametrize(
    'nav_shape', (
        None,
        (14, 14),
        (47, 13),
        "random1d",
        "random2d",
        "random3d",
    ),
)
@pytest.mark.parametrize(
    'sync_offset', (
        0, 1, -1, -10, 13, 13*15, -13*14,
        "random",
        -13*17+1,
        13*17-1,
    ),
)
@pytest.mark.parametrize(
    'sig_shape', (
        None, (24, 19), (24*19, )
    ),
)
@pytest.mark.parametrize(
    'use_roi', (False, True)
)
def test_reshape_sync_offset(
        raw_csr_generated, mock_sparse_data, lt_ctx, sync_offset, nav_shape, sig_shape, use_roi):
    if sync_offset == 'random':
        sync_offset = np.random.randint(low=-13*17+1, high=13*17)
    if nav_shape == 'random1d':
        nav_shape = (np.random.randint(low=1, high=13*18),)
    elif nav_shape == 'random2d':
        nav_shape = (
            np.random.randint(low=1, high=13*18),
            np.random.randint(low=1, high=13*18),
        )
    elif nav_shape == 'random3d':
        nav_shape = (
            np.random.randint(low=1, high=13*18),
            np.random.randint(low=1, high=13*18),
            np.random.randint(low=1, high=13*18),
        )

    orig, data_flat = mock_sparse_data
    data = data_flat.reshape(raw_csr_generated.shape)
    # Otherwise memory and raw_csr use different approach to determine shape
    # that yields different results
    if nav_shape is None and sig_shape is not None:
        mem_nav_shape = raw_csr_generated.shape.nav
    else:
        mem_nav_shape = nav_shape
    if sig_shape is None and nav_shape is not None:
        mem_sig_shape = raw_csr_generated.shape.sig
    else:
        mem_sig_shape = sig_shape
    ref_ds = lt_ctx.load(
        'memory',
        data=data,
        nav_shape=mem_nav_shape,
        sig_shape=mem_sig_shape,
        sync_offset=sync_offset,
    )

    if use_roi:
        roi = np.random.choice([True, False], size=ref_ds.shape.nav)
    else:
        roi = None

    print('nav_shape', nav_shape)
    print('sig_shape', sig_shape)
    print('sync_offset', sync_offset)
    print('roi', roi)

    ds = lt_ctx.load(
        'raw_csr', path=raw_csr_generated._path,
        sync_offset=sync_offset, nav_shape=nav_shape, sig_shape=sig_shape
    )

    assert tuple(ref_ds.shape.sig) == tuple(ds.shape.sig)
    assert tuple(ref_ds.shape.nav) == tuple(ds.shape.nav)

    masks = np.random.random(size=(3, *ref_ds.shape.sig))
    udf_masks = ApplyMasksUDF(mask_factories=lambda: masks)
    udf_std = StdDevUDF()

    ref_result = lt_ctx.run_udf(udf=(udf_masks, udf_std), dataset=ref_ds)
    result = lt_ctx.run_udf(udf=(udf_masks, udf_std), dataset=ds)

    r1 = ref_result[0]['intensity'].raw_data
    r2 = result[0]['intensity'].raw_data

    print(
        np.max((r1 - r2) / np.maximum(0.00001, (np.abs(r1) + np.abs(r2))))
    )

    assert_allclose(
        ref_result[0]['intensity'].raw_data,
        result[0]['intensity'].raw_data,
        rtol=1e-5,
        atol=1e-8,
    )
    assert_allclose(
        ref_result[1]['std'].raw_data,
        result[1]['std'].raw_data,
        rtol=1e-5,
        atol=1e-8,
    )
    assert_allclose(
        ref_result[1]['num_frames'].raw_data,
        result[1]['num_frames'].raw_data,
        rtol=1e-5,
        atol=1e-8,
    )


def test_uint64ptr(lt_ctx_fast, raw_csr_generated_uint64):
    prime_numba_cache(raw_csr_generated_uint64)


def test_large_file_detect(monkeypatch, default_raw, inline_executor_fast):
    # Use a mock load_toml function to check if we loaded a large file
    # could do this with a real mock object...
    load_called = False

    def mock_load_toml(*args, **kwargs):
        nonlocal load_called
        load_called = True

    monkeypatch.setattr(raw_csr, "load_toml", mock_load_toml)

    # default_raw is a 16 MB file
    filepath = default_raw._path
    detects = RawCSRDataSet.detect_params(filepath, inline_executor_fast)
    assert not detects
    assert not load_called



from typing import Optional
from libertem.web.notebook_generator.template import TemplateBase


class GeneratorHelper(TemplateBase):
    """
    Helper class for generating code corresponding to analysis.

    code for creating and running analysis, saving the result,
    plotting the result, documentation, specific dependencies
    are generated.
    """

    short_name: Optional[str] = None
    api: Optional[str] = None

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        """
        Get analysis dependencies.
        """
        return None

    def convert_params(self):
        """
        Format analysis parameters.
        """
        return None

    def get_plot(self):
        """
        Get code for ploting analysis.
        """
        return None

    def get_docs(self):
        """
        Get documentation for analysis.
        """
        return None

    def get_save(self):
        '''
        Get code for saving result.
        '''
        data = {'short': self.short_name}
        save = self.format_template(self.temp_save, data)
        return save

    def format_docs(self, title, docs_rst):
        """
        function to format docs for notebook
        """
        docs = f'# {title}\n\n<pre>{docs_rst}</pre>'
        return docs

    def get_roi(self):
        """
        Get code for roi
        """
        roi = ["nx, ny = ds.shape.nav"]
        params = self.params['roi']
        if params == {}:
            roi = ["roi = None"]

        elif params["shape"] == "disk":
            x = params["cx"]
            y = params["cy"]
            r = params["r"]
            p = f"centerX={x}, centerY={y}, imageSizeX=nx, imageSizeY=ny, radius={r}"
            roi.append(f"roi = masks.circular({p})")

        elif params["shape"] == "rect":
            x = params['x']
            y = params['y']
            width = params['width']
            height = params['height']
            p = f"X={x}, Y={y}, Width={width}, Height={height}, imageSizeX=nx, imageSizeY=ny"
            roi.append(f"roi = masks.rectangular({p})")

        return '\n'.join(roi)

    def get_analysis(self):
        '''
        get code corresponding to create and run analysis.
        override for adding specific code for analysis
        subclasses.
        '''
        params_ = self.convert_params()

        data = {
            'short': self.short_name,
            'analysis_api': self.api,
            'params': params_,
        }

        analy_ = self.format_template(self.temp_analysis, data)

        return analy_



import inspect
import pytest


from libertem.io.dataset import filetypes, get_dataset_cls


@pytest.mark.parametrize(
    "ds_key", tuple(filetypes.keys())
)
def test_missing(lt_ctx, ds_key):
    ds_class = get_dataset_cls(ds_key)
    classname = ds_class.__name__
    params = inspect.signature(ds_class.__init__).parameters
    for extension in ds_class.get_supported_extensions():
        filename = f"not_a_file.{extension}"
        with pytest.raises(Exception) as e:
            lt_ctx.load(
                ds_key, filename,
            )
        e_msg = str(e.value)

        print(f'{classname} + {filename} => {e.typename}("{e_msg}")')

        if "nav_shape" in e_msg and "nav_shape" in params:
            nav_shape = (8, 8)
            with pytest.raises(Exception) as e:
                lt_ctx.load(
                    ds_key, filename, nav_shape=nav_shape
                )
            e_msg = str(e.value)
            print(f'{classname} + {filename} + nav_shape={nav_shape} => {e.typename}("{e_msg}")')



import functools

import numpy as np

from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


def test_auto(lt_ctx):
    data = _mk_random((16, 8, 32, 64))

    dataset = MemoryDataSet(data=data, tileshape=(8, 32, 64),
                            num_partitions=2, sig_dims=2)

    auto_result = lt_ctx.map(dataset=dataset, f=functools.partial(np.sum, axis=-1))

    naive_result = data.sum(axis=-1)

    assert np.allclose(auto_result.data, naive_result)


def test_auto_weird(lt_ctx):
    data = _mk_random((16, 8, 32, 64))

    dataset = MemoryDataSet(data=data, tileshape=(8, 32, 64),
                            num_partitions=2, sig_dims=2)

    def f(frame):
        return [
            "Shape %s" % str(frame.shape),
            dict(shape=frame.shape, sum=frame.sum()),
            lambda x: x,
            MemoryDataSet
        ]

    auto_result = lt_ctx.map(dataset=dataset, f=f)
    item = auto_result.data[0, 0]

    assert len(item) == 4
    assert isinstance(item[0], str)
    assert isinstance(item[1], dict)
    assert callable(item[2])
    assert item[2](1) == 1
    assert isinstance(item[3], type)



import numpy as np
import pytest
from unittest import mock

import IPython

from libertem.viz.bqp import BQLive2DPlot
from libertem.udf.sum import SumUDF
from libertem.udf.sumsigudf import SumSigUDF


def test_bqp_smoke(monkeypatch, lt_ctx, default_raw):
    pytest.importorskip("bqplot")
    udf = SumUDF()
    monkeypatch.setattr(IPython.display, 'display', mock.MagicMock())
    monkeypatch.setattr(lt_ctx, 'plot_class', BQLive2DPlot)
    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=True)
    IPython.display.display.assert_called()


def test_bqp_asymm(monkeypatch, lt_ctx, default_raw_asymm):
    pytest.importorskip("bqplot")
    udfs = [SumSigUDF(), SumUDF()]
    monkeypatch.setattr(IPython.display, 'display', mock.MagicMock())
    monkeypatch.setattr(lt_ctx, 'plot_class', BQLive2DPlot)
    lt_ctx.run_udf(dataset=default_raw_asymm, udf=udfs, plots=True)
    IPython.display.display.assert_called()


def test_empty(monkeypatch, lt_ctx, default_raw_asymm):
    default_raw = default_raw_asymm
    pytest.importorskip("bqplot")
    udf = SumSigUDF()
    monkeypatch.setattr(IPython.display, 'display', mock.MagicMock())
    monkeypatch.setattr(lt_ctx, 'plot_class', BQLive2DPlot)

    def extract(udf_result, damage):
        return udf_result['intensity'].data, np.zeros_like(damage)

    plot = BQLive2DPlot(dataset=default_raw, udf=udf, channel=extract)

    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=[plot])



import os
import sys
import json
import random

import numpy as np
import pytest
import cloudpickle

from libertem.analysis.raw import PickFrameAnalysis
from libertem.executor.inline import InlineJobExecutor
from libertem.io.dataset.blo import BloDataSet
from libertem.io.dataset.base import TilingScheme, BufferedBackend, MMapBackend, DirectBackend
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.raw import PickUDF

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

try:
    import rsciio
    import rsciio.blockfile
except ModuleNotFoundError:
    rsciio = None


BLO_TESTDATA_PATH = os.path.join(get_testdata_path(), 'default.blo')
HAVE_BLO_TESTDATA = os.path.exists(BLO_TESTDATA_PATH)

pytestmark = pytest.mark.skipif(not HAVE_BLO_TESTDATA, reason="need .blo testdata")  # NOQA


@pytest.fixture()
def default_blo(lt_ctx):
    ds = lt_ctx.load(
        "blo",
        path=str(BLO_TESTDATA_PATH),
        io_backend=MMapBackend(),
    )
    return ds


@pytest.fixture
def buffered_blo(lt_ctx):
    buffered = BufferedBackend()
    ds_buffered = lt_ctx.load(
        "blo",
        path=str(BLO_TESTDATA_PATH),
        io_backend=buffered,
    )
    return ds_buffered


@pytest.fixture
def direct_blo(lt_ctx):
    direct = DirectBackend()
    ds_buffered = lt_ctx.load(
        "blo",
        path=str(BLO_TESTDATA_PATH),
        io_backend=direct,
    )
    return ds_buffered


@pytest.fixture(scope='module')
def default_blo_raw():
    res = rsciio.blockfile.file_reader(str(BLO_TESTDATA_PATH), lazy=True)
    return res[0]['data']


def test_simple_open(default_blo):
    assert tuple(default_blo.shape) == (90, 121, 144, 144)


def test_check_valid(default_blo):
    assert default_blo.check_valid()


def test_detect():
    assert BloDataSet.detect_params(
        path=str(BLO_TESTDATA_PATH),
        executor=InlineJobExecutor()
    )["parameters"]


def test_read(default_blo):
    partitions = default_blo.get_partitions()
    p = next(partitions)
    # FIXME: partition shape can vary by number of cores
    # assert tuple(p.shape) == (11, 121, 144, 144)
    tileshape = Shape(
        (8,) + tuple(default_blo.shape.sig),
        sig_dims=default_blo.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_blo.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == (8, 144, 144)
    for _ in tiles:
        pass


def test_scheme_too_large(default_blo):
    partitions = default_blo.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_blo.shape.sig),
        sig_dims=default_blo.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_blo.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == (depth, 144, 144)
    for _ in tiles:
        pass


@pytest.mark.skipif(rsciio is None, reason="No rosettasciio found")
def test_comparison(default_blo, default_blo_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_blo_raw, (-1, *tuple(default_blo.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_blo)


@pytest.mark.skipif(rsciio is None, reason="No rosettasciio found")
def test_comparison_roi(default_blo, default_blo_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_blo.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_blo_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_blo, roi=roi)


def test_pickle_meta_is_small(default_blo):
    pickled = cloudpickle.dumps(default_blo._meta)
    cloudpickle.loads(pickled)
    assert len(pickled) < 512


def test_pickle_fileset_is_small(default_blo):
    pickled = cloudpickle.dumps(default_blo._get_fileset())
    cloudpickle.loads(pickled)
    assert len(pickled) < 1024


def test_apply_mask_analysis(default_blo, lt_ctx):
    mask = np.ones((144, 144))
    analysis = lt_ctx.create_mask_analysis(factories=[lambda: mask], dataset=default_blo)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (90, 121)


def test_sum_analysis(default_blo, lt_ctx):
    analysis = lt_ctx.create_sum_analysis(dataset=default_blo)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (144, 144)


def test_pick_analysis(default_blo, lt_ctx):
    analysis = PickFrameAnalysis(dataset=default_blo, parameters={"x": 16, "y": 16})
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (144, 144)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.slow
def test_correction(default_blo, lt_ctx, with_roi):
    ds = default_blo

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx, exclude=[(55, 92), (61, 31)])
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_cache_key_json_serializable(default_blo):
    json.dumps(default_blo.get_cache_key())


@pytest.mark.dist
def test_blo_dist(dist_ctx):
    ds = BloDataSet(path=BLO_TESTDATA_PATH)
    ds = ds.initialize(dist_ctx.executor)
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    results = dist_ctx.run(analysis)
    assert results[0].raw_data.shape == (144, 144)


def test_positive_sync_offset(lt_ctx):
    udf = SumSigUDF()
    sync_offset = 2

    ds = lt_ctx.load(
        "blo", path=BLO_TESTDATA_PATH, nav_shape=(4, 2),
    )

    result = lt_ctx.run_udf(dataset=ds, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    ds_with_offset = lt_ctx.load(
        "blo", path=BLO_TESTDATA_PATH, nav_shape=(4, 2), sync_offset=sync_offset
    )

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.shape.nav.size - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


def test_negative_sync_offset(lt_ctx):
    udf = SumSigUDF()
    sync_offset = -2

    ds = lt_ctx.load(
        "blo", path=BLO_TESTDATA_PATH, nav_shape=(4, 2),
    )

    result = lt_ctx.run_udf(dataset=ds, udf=udf)
    result = result['intensity'].raw_data[:ds._meta.shape.nav.size - abs(sync_offset)]

    ds_with_offset = lt_ctx.load(
        "blo", path=BLO_TESTDATA_PATH, nav_shape=(4, 2), sync_offset=sync_offset
    )

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


def test_offset_smaller_than_image_count(lt_ctx):
    sync_offset = -10900

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "blo",
            path=BLO_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-10890, 10890\), which is \(-image_count, image_count\)"
    )


def test_offset_greater_than_image_count(lt_ctx):
    sync_offset = 10900

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "blo",
            path=BLO_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-10890, 10890\), which is \(-image_count, image_count\)"
    )


def test_reshape_nav(lt_ctx):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("blo", path=BLO_TESTDATA_PATH, nav_shape=(8,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("blo", path=BLO_TESTDATA_PATH, nav_shape=(4, 2))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("blo", path=BLO_TESTDATA_PATH, nav_shape=(2, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


def test_incorrect_sig_shape(lt_ctx):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "blo",
            path=BLO_TESTDATA_PATH,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 20736"
    )


def test_compare_backends(lt_ctx, default_blo, buffered_blo):
    y = random.choice(range(default_blo.shape.nav[0]))
    x = random.choice(range(default_blo.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_blo,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_blo,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_blo, direct_blo):
    y = random.choice(range(default_blo.shape.nav[0]))
    x = random.choice(range(default_blo.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_blo,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_blo,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_blo, buffered_blo, as_sparse):
    roi = np.zeros(default_blo.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[16] = True
    roi[32] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_blo, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_blo, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("blo", BLO_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



import logging

import numba
from numba.typed import List as NumbaList
import numpy as np
from sparseconverter import check_shape

from libertem.common.numba import numba_ravel_multi_index_single as _ravel_multi_index, cached_njit


log = logging.getLogger(__name__)


@numba.njit(inline='always')
def _default_px_to_bytes(
    bpp, frame_in_file_idx, slice_sig_size, sig_size, sig_origin,
    frame_footer_bytes, frame_header_bytes, file_header_bytes,
    file_idx, read_ranges,
):
    """
    Convert from a slice (sig_origin, slice_sig_size) to a byte slice (start, stop)
    and append the result to the read_ranges numba List
    """
    # we are reading a part of a single frame, so we first need to find
    # the offset caused by headers and footers:
    footer_offset = frame_footer_bytes * frame_in_file_idx
    header_offset = frame_header_bytes * (frame_in_file_idx + 1)
    byte_offset = file_header_bytes + footer_offset + header_offset

    # now let's figure in the current frame index:
    # (go down into the file by full frames; `sig_size`)
    offset = byte_offset + frame_in_file_idx * sig_size * bpp // 8

    # offset in px in the current frame:
    sig_origin_bytes = sig_origin * bpp // 8

    start = offset + sig_origin_bytes

    # size of the sig part of the slice:
    sig_size_bytes = slice_sig_size * bpp // 8

    stop = start + sig_size_bytes

    read_ranges.append((file_idx, start, stop))


@numba.njit(boundscheck=True, nogil=True)
def _find_file_for_frame_idx(fileset_arr, frame_idx):
    """
    Find the file in `fileset_arr` that contains
    `frame_idx` and return its index using binary search.

    Worst case: something like 2**20 files, each containing
    a single frame.

    `fileset_arr` is an array of shape (number_files, 4)
    where `fileset_arr[i]` is:

        (start_idx, end_idx, file_header_size, file_idx)

    It must be sorted by `start_idx` and the defined intervals must not overlap.
    """
    while True:
        num_files = fileset_arr.shape[0]
        mid = num_files // 2
        mid_file = fileset_arr[mid]

        if mid_file[0] <= frame_idx and mid_file[1] > frame_idx:
            return mid_file[2]
        elif mid_file[0] > frame_idx:
            fileset_arr = fileset_arr[:mid]
        else:
            fileset_arr = fileset_arr[mid + 1:]


@numba.njit(inline='always')
def _default_read_ranges_tile_block(
    slices_arr, fileset_arr, slice_sig_sizes, sig_origins,
    inner_indices_start, inner_indices_stop, frame_indices, sig_size,
    px_to_bytes, bpp, frame_header_bytes, frame_footer_bytes, file_idxs,
    slice_offset, extra, sig_shape,
):
    result = NumbaList()

    # positions in the signal dimensions:
    for slice_idx in range(slices_arr.shape[0]):
        # (offset, size) arrays defining what data to read (in pixels)
        # NOTE: assumes contiguous tiling scheme
        # (i.e. a shape like (1, 1, ..., 1, X1, ..., XN))
        # where X1 is <= the dataset shape at that index, and X2, ..., XN are
        # equal to the dataset shape at that index
        slice_origin = slices_arr[slice_idx][0]
        slice_shape = slices_arr[slice_idx][1]
        slice_sig_size = slice_sig_sizes[slice_idx]
        sig_origin = sig_origins[slice_idx]

        read_ranges = NumbaList()

        # inner "depth" loop along the (flat) navigation axis of a tile:
        for i, inner_frame_idx in enumerate(range(inner_indices_start, inner_indices_stop)):
            inner_frame = frame_indices[inner_frame_idx]

            file_idx = file_idxs[i]
            f = fileset_arr[file_idx]

            frame_in_file_idx = inner_frame - f[0]
            file_header_bytes = f[3]

            # px_to_bytes is the format-specific translation of pixel
            # coordinates (slice_sig_size, sig_size, sig_origin)
            # to bytes, which are appended as tuples (file_idx, start, stop)
            # to the `read_ranges` list.
            px_to_bytes(
                bpp=bpp,
                frame_in_file_idx=frame_in_file_idx,
                slice_sig_size=slice_sig_size,
                sig_size=sig_size,
                sig_origin=sig_origin,
                frame_footer_bytes=frame_footer_bytes,
                frame_header_bytes=frame_header_bytes,
                file_header_bytes=file_header_bytes,
                file_idx=file_idx,
                read_ranges=read_ranges,
            )

        # the indices are compressed to the selected frames
        compressed_slice = np.array([
            [slice_offset + inner_indices_start] + [i for i in slice_origin],
            [inner_indices_stop - inner_indices_start] + [i for i in slice_shape],
        ])
        result.append((slice_idx, compressed_slice, read_ranges))

    return result


def make_get_read_ranges(
    px_to_bytes=_default_px_to_bytes,
    read_ranges_tile_block=_default_read_ranges_tile_block,
):
    """
    Translate the `TilingScheme` combined with the `roi` into (pixel)-read-ranges,
    together with their tile slices.

    Parameters
    ----------

    start_at_frame
        Dataset-global first frame index to read

    stop_before_frame
        Stop before this frame index

    tiling_scheme
        Description on how the data should be tiled

    fileset_arr
        Array of shape (number_of_files, 3) where the last dimension contains
        the following values: `(start_idx, end_idx, file_idx)`, where
        `[start_idx, end_idx)` defines which frame indices are contained
        in the file.

    roi
        Region of interest (for the full dataset)

    bpp : int
        Bits per pixel, including padding

    Returns
    -------

    (tile_slice, read_ranges)
        read_ranges is an ndarray with shape (number_of_tiles, depth, 3)
        where the last dimension contains: file index, start_byte, stop_byte
    """

    @cached_njit(boundscheck=True, cache=True, nogil=True)
    def _get_read_ranges_inner(
        start_at_frame, stop_before_frame, roi_nonzero, depth,
        slices_arr, fileset_arr, sig_shape,
        bpp, sync_offset=0, extra=None, frame_header_bytes=0, frame_footer_bytes=0,
    ):
        result = NumbaList()
        # Use NumPy prod for Numba compilation
        sig_size = np.prod(np.array(sig_shape).astype(np.int64))

        if roi_nonzero is None:
            frame_indices = np.arange(max(0, start_at_frame), stop_before_frame)
            # in case of a negative sync_offset, start_at_frame can be negative
            if start_at_frame < 0:
                slice_offset = abs(sync_offset)
            else:
                slice_offset = start_at_frame - sync_offset
        else:
            shifted_roi = roi_nonzero + sync_offset
            roi_mask = np.logical_and(shifted_roi >= max(0, start_at_frame),
                                      shifted_roi < stop_before_frame)
            frame_indices = shifted_roi[roi_mask]

            if start_at_frame < 0:
                slice_offset = np.sum(roi_nonzero < abs(sync_offset))
            else:
                slice_offset = np.sum(roi_nonzero < start_at_frame - sync_offset)

        num_indices = frame_indices.shape[0]

        # indices into `frame_indices`:
        inner_indices_start = 0
        inner_indices_stop = min(depth, num_indices)

        # this should be `prod(..., axis=-1)``, which is not supported by numba yet:
        # slices that divide the signal dimensions:
        slice_sig_sizes = np.array([
            # Use NumPy prod for Numba compilation
            np.prod(slices_arr[slice_idx, 1, :].astype(np.int64))
            for slice_idx in range(slices_arr.shape[0])
        ])

        sig_origins = np.array([
            _ravel_multi_index(slices_arr[slice_idx][0], sig_shape)
            for slice_idx in range(slices_arr.shape[0])
        ])

        # outer "depth" loop skipping over `depth` frames at a time:
        while inner_indices_start < num_indices:
            file_idxs = np.array([
                _find_file_for_frame_idx(fileset_arr, frame_indices[inner_frame_idx])
                for inner_frame_idx in range(inner_indices_start, inner_indices_stop)
            ])

            for slice_idx, compressed_slice, read_ranges in read_ranges_tile_block(
                slices_arr, fileset_arr, slice_sig_sizes, sig_origins,
                inner_indices_start, inner_indices_stop, frame_indices, sig_size,
                px_to_bytes, bpp, frame_header_bytes, frame_footer_bytes, file_idxs,
                slice_offset, extra=extra, sig_shape=sig_shape,
            ):
                result.append((compressed_slice, read_ranges, slice_idx))

            inner_indices_start = inner_indices_start + depth
            inner_indices_stop = min(inner_indices_stop + depth, num_indices)

        result_slices = np.zeros((len(result), 2, 1 + len(sig_shape)), dtype=np.int64)
        for tile_idx, res in enumerate(result):
            result_slices[tile_idx] = res[0]

        if len(result) == 0:
            return (
                result_slices,
                np.zeros((len(result), depth, 3), dtype=np.int64),
                np.zeros((len(result)), dtype=np.int64),
            )

        lengths = [len(res[1])
                   for res in result]
        max_rr_per_tile = max(lengths)

        slice_indices = np.zeros(len(result), dtype=np.int64)

        # read_ranges_tile_block can decide how many entries there are per read range,
        # so we need to generate a result array with the correct size:
        rr_num_entries = max(3, len(result[0][1][0]))
        result_ranges = np.zeros((len(result), max_rr_per_tile, rr_num_entries), dtype=np.int64)
        for tile_idx, res in enumerate(result):
            for depth_idx, read_range in enumerate(res[1]):
                result_ranges[tile_idx][depth_idx] = read_range
            slice_indices[tile_idx] = res[2]

        return result_slices, result_ranges, slice_indices
    return _get_read_ranges_inner


default_get_read_ranges = make_get_read_ranges()


class DataTile:
    def __init__(self, data, tile_slice, scheme_idx):
        if isinstance(data, DataTile):
            data = data.data
        check_shape(data, tile_slice.shape)
        self._data = data
        self.tile_slice = tile_slice
        self.scheme_idx = scheme_idx

    @property
    def flat_data(self) -> np.ndarray:
        """
        Flatten the data.

        The result is a 2D array where each row contains pixel data
        from a single frame. It is just a reshape, so it is a view into
        the original data.
        """
        shape = self.tile_slice.shape
        tileshape = (
            shape.nav.size,    # stackheight, number of frames we process at once
            shape.sig.size,    # framesize, number of pixels per tile
        )
        return self._data.reshape(tileshape)

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def shape(self):
        return tuple(self.tile_slice.shape)

    @property
    def size(self):
        return self.tile_slice.shape.size

    @property
    def data(self):
        return self._data

    @property
    def c_contiguous(self):
        try:
            return self._data.flags.c_contiguous
        except AttributeError:
            return None

    def __repr__(self):
        return "<DataTile %r scheme_idx=%d>" % (self.tile_slice, self.scheme_idx)



import pytest

import dask.array as da
import numpy as np

from libertem.executor.utils.dask_inplace import DaskInplaceWrapper
from utils import _mk_random


def get_wrapped_data(shape, dtype):
    data = _mk_random(shape, dtype=dtype)
    data_dask = da.from_array(data.copy())
    dask_wrapped = DaskInplaceWrapper(data_dask)
    return data, dask_wrapped


def test_inplace_methods():
    shape = (16, 8, 32, 64)
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    assert dask_wrapped.flags.c_contiguous
    assert np.allclose(dask_wrapped.data.compute(), data)
    assert np.allclose(dask_wrapped.unwrap_sliced().compute(), data)
    assert dask_wrapped.shape == data.shape
    assert dask_wrapped.dtype == dtype
    assert dask_wrapped.size == data.size


def test_inplace_raw():
    shape = (16, 8, 32, 64)
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)
    sl = np.s_[4, :, :, :]

    assert np.allclose(dask_wrapped[sl].compute(), data[sl])
    dask_wrapped[sl] = 55.
    data[sl] = 55.
    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_get_with_slice(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4, :, :, :]
    subslice = np.s_[5, 0, 4]

    dask_wrapped.set_slice(sl)
    assert np.allclose(dask_wrapped.unwrap_sliced().compute(), data[sl])
    assert dask_wrapped[subslice].compute() == data[sl][subslice]


@pytest.mark.parametrize(
    "shape", ((8, 8),))
def test_inplace_set_with_ints(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4, :]
    subslice = np.s_[5]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)

    dask_wrapped.clear_slice()
    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 64),))
def test_inplace_set_with_ints2(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[:, 2, :]
    subslice = np.s_[3, 4]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 16),))
def test_inplace_set_with_neg_ints(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[-3]
    subslice = np.s_[-7]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_set_with_colon(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4, :, :, :]

    dask_wrapped.set_slice(sl)
    dask_wrapped[:] = 55.
    data[sl][:] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_set_with_ranges(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4:7, :, :, :]
    subslice = np.s_[3:6, 10:22, :]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_set_with_ranges2(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[:]
    subslice = np.s_[3:6, 10:22, :]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_set_with_ellipsis(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4:7, ...]
    subslice = np.s_[3:6, 10:22, :]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_set_with_ellipsis2(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4:7, ...]
    subslice = np.s_[3:6, ...]

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = 55.
    data[sl][subslice] = 55.

    assert np.allclose(dask_wrapped.data.compute(), data)


def test_inplace_set_with_ellipsis3():
    _, dask_wrapped = get_wrapped_data((16, 8, 32, 64), np.float32)

    sl = np.s_[4:7, ...]
    # No support for this yet !
    subslice = np.s_[..., 3:6]

    dask_wrapped.set_slice(sl)
    with pytest.raises(IndexError):
        dask_wrapped[subslice] = 55.


@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),))
def test_inplace_set_with_array(shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = np.s_[4]
    subslice = np.s_[3:6, 10:22, :]

    target_shape = data[sl][subslice].shape
    set_values = np.random.random(size=target_shape)

    dask_wrapped.set_slice(sl)
    dask_wrapped[subslice] = set_values
    data[sl][subslice] = set_values

    assert np.allclose(dask_wrapped.data.compute(), data)


def int_for_dim(dim):
    return np.random.randint(-dim, dim)


def slice_range_for_dim(dim):
    start = int_for_dim(dim)
    stop = np.random.randint(start, dim)
    possible_steps = [None, 1]
    if stop - start > 1:
        # This avoids a bug in Dask where if step > (stop - start)
        # it is computed to be a zero-length slice, raising a ValueError
        # whereas numpy does a no-op
        possible_steps += [2]
    step = np.random.choice(possible_steps)
    return slice(start, stop, step)


def null_slice(dim=None):
    return slice(None, None, None)


def ellipsis(dim=None):
    return Ellipsis


def random_slice(shape, final_ellipsis=False):
    slices = tuple()
    has_ellipsis = False
    for dim in shape:
        if np.random.choice([True, False], p=[0.2, 0.8]):
            break
        choices = [null_slice]
        if dim > 1:
            choices += [int_for_dim, slice_range_for_dim]
        if not has_ellipsis:
            choices += [ellipsis]
        slicetype = np.random.choice(choices)
        slices = slices + (slicetype(dim),)
        # Handle double Ellipsis case
        if slices[-1] is Ellipsis:
            has_ellipsis = True
            if final_ellipsis:
                break
    if len(slices) == 0:
        slices = null_slice()
    return slices


@pytest.mark.parametrize(
    "repeat_number", range(20))
@pytest.mark.parametrize(
    "shape", ((16, 8, 32, 64),
              (16, 16, 8),
              (16, 16),
              (16,),))
def test_random_set_with_array(repeat_number, shape):
    dtype = np.float32
    data, dask_wrapped = get_wrapped_data(shape, dtype)

    sl = random_slice(shape)

    # Encountered a rare bug where my generated slice would
    # be invalid for numpy slicing for an unknown reason
    # as an out of range integer for dimension
    try:
        slice_into = data[sl]
    except IndexError:
        print(f'Skipping due to invalid primary slice {sl} on shape {shape}')
        return
    if np.isscalar(slice_into):
        subslice = None
    else:
        subslice = random_slice(slice_into.shape, final_ellipsis=True)

    # Idem
    try:
        target_shape = data[sl][subslice].shape
    except IndexError:
        print(f'Skipping due to invalid secondary slice {subslice} '
              f'on shape {shape} with sl {sl}')
        return

    set_values = np.random.random(size=target_shape)
    dask_wrapped.set_slice(sl)

    # Assignment to an empty slice sometimes raises a ValueError in Dask
    # but is a no-op in Numpy. The 'sometimes' refers to specific combinations
    # of zero-length, [:] or Ellipsis subslices not raising an error
    try:
        dask_wrapped[subslice] = set_values
    except ValueError:
        assert 0 in target_shape or 1 in target_shape
        # Can't find consistent behaviour in Dask so better to return to avoid
        # tests failing for things outside of our control
        return

    if set_values.size == 1:
        set_values = set_values.item()

    # Strange-ish case of Dask supporting an assignment but Numpy raises!
    if np.isscalar(slice_into):
        # Would raise TypeError if using subslice even if None
        data[sl] = set_values
    else:
        data[sl][subslice] = set_values
    assert np.allclose(dask_wrapped.data.compute(), data)



import os
import sys
from glob import glob
import random

import numpy as np
import pytest

from libertem.io.dataset.dm import DMDataSet
from libertem.udf.sum import SumUDF
from libertem.common import Shape
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.raw import PickUDF
from libertem.io.dataset.base import (
    TilingScheme, BufferedBackend, MMapBackend, DataSetException, DirectBackend
)

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

try:
    import hyperspy.api as hs
except ModuleNotFoundError:
    hs = None

DM_TESTDATA_PATH = os.path.join(get_testdata_path(), 'dm')
HAVE_DM_TESTDATA = os.path.exists(DM_TESTDATA_PATH)

pytestmark = pytest.mark.skipif(not HAVE_DM_TESTDATA, reason="need .dm4 testdata")  # NOQA


@pytest.fixture(scope='module')
def dm_stack_glob():
    return glob(os.path.join(DM_TESTDATA_PATH, '2018-7-17*.dm4'))


@pytest.fixture(scope='module')
def dm_3d_glob():
    return glob(os.path.join(DM_TESTDATA_PATH, '3D', '*.dm3'))


@pytest.fixture
def default_dm(lt_ctx, dm_stack_glob):
    files = list(sorted(dm_stack_glob))
    ds = lt_ctx.load(
        "dm",
        files=files,
        io_backend=MMapBackend(),
    )
    return ds


@pytest.fixture(scope='module')
def default_dm_raw(dm_stack_glob):
    files = list(sorted(dm_stack_glob))
    return np.stack([hs.load(file).data for file in files])


@pytest.fixture
def buffered_dm(lt_ctx, dm_stack_glob):
    buffered = BufferedBackend()
    files = list(sorted(dm_stack_glob))
    return lt_ctx.load(
        "dm",
        files=files,
        io_backend=buffered,
    )


@pytest.fixture
def direct_dm(lt_ctx, dm_stack_glob):
    buffered = DirectBackend(dm_stack_glob)
    files = list(sorted(dm_stack_glob))
    return lt_ctx.load(
        "dm",
        files=files,
        io_backend=buffered,
    )


@pytest.fixture
def dm_stack_of_3d(lt_ctx, dm_3d_glob):
    files = list(sorted(dm_3d_glob))
    ds = lt_ctx.load("dm", files=files)
    return ds


@pytest.fixture(scope='module')
def default_dm_3d_raw(dm_3d_glob):
    files = list(sorted(dm_3d_glob))
    return np.concatenate([hs.load(file).data for file in files], axis=0)


def test_simple_open(default_dm):
    assert tuple(default_dm.shape) == (10, 3838, 3710)


def test_check_valid(default_dm):
    default_dm.check_valid()


@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
def test_comparison(default_dm, default_dm_raw, lt_ctx_fast):
    udf = ValidationUDF(reference=default_dm_raw)
    lt_ctx_fast.run_udf(udf=udf, dataset=default_dm)


@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
def test_comparison_roi(default_dm, default_dm_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_dm.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_dm_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_dm, roi=roi)


@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
def test_comparison_3d(dm_stack_of_3d, default_dm_3d_raw, lt_ctx_fast):
    udf = ValidationUDF(reference=default_dm_3d_raw)
    lt_ctx_fast.run_udf(udf=udf, dataset=dm_stack_of_3d)


@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
def test_comparison_3d_roi(dm_stack_of_3d, default_dm_3d_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(dm_stack_of_3d.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_dm_3d_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=dm_stack_of_3d, roi=roi)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.slow
def test_correction(default_dm, lt_ctx, with_roi):
    ds = default_dm

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)

# Disabled while DMDataSet.detect_params returns False
# to disable it in the Web UI
# def test_detect_1(lt_ctx):
#     fpath = os.path.join(DM_TESTDATA_PATH, '2018-7-17 15_29_0000.dm4')
#     assert DMDataSet.detect_params(
#         path=fpath,
#         executor=lt_ctx.executor,
#     )["parameters"] == {
#         'files': [fpath],
#     }


def test_get_metadata():
    import distributed
    from libertem.api import Context
    from libertem.io.dataset.dm import _get_metadata

    with distributed.Client(
            n_workers=2,
            threads_per_worker=1,
            processes=False
    ) as _:
        ctx = Context.make_with("dask-integration")
        fpath = os.path.join(DM_TESTDATA_PATH, '2018-7-17 15_29_0000.dm4')
        ctx.executor.run_function(lambda: _get_metadata(fpath))


def test_get_metadata_2():
    from libertem.io.dataset.dm import _get_metadata

    fpath = os.path.join(DM_TESTDATA_PATH, '2018-7-17 15_29_0000.dm4')
    _get_metadata(fpath)


def test_detect_2(lt_ctx):
    assert DMDataSet.detect_params(
        path="nofile.someext",
        executor=lt_ctx.executor,
    ) is False


def test_same_offset(lt_ctx, dm_stack_glob):
    files = dm_stack_glob
    ds = lt_ctx.load("dm", files=files, same_offset=True)
    ds.check_valid()


def test_repr(default_dm):
    assert repr(default_dm) == "<DMDataSet for a stack of 10 files>"


@pytest.mark.dist
def test_dm_dist(dist_ctx, dm_stack_glob):
    files = dist_ctx.executor.run_function(lambda: list(sorted(dm_stack_glob)))
    print(files)
    ds = DMDataSet(files=files)
    ds = ds.initialize(dist_ctx.executor)
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    roi = np.random.choice([True, False], size=len(files))
    results = dist_ctx.run(analysis, roi=roi)
    assert results[0].raw_data.shape == (3838, 3710)


def test_dm_stack_fileset_offsets(dm_stack_of_3d, lt_ctx):
    fs = dm_stack_of_3d._get_fileset()
    f0, f1 = fs

    # check that the offsets in the fileset are correct:
    assert f0.num_frames == 20
    assert f1.num_frames == 20
    assert f0.start_idx == 0
    assert f0.end_idx == 20
    assert f1.start_idx == 20
    assert f1.end_idx == 40

    lt_ctx.run_udf(dataset=dm_stack_of_3d, udf=SumUDF())


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_positive_sync_offset(lt_ctx, io_backend, dm_stack_glob):
    udf = SumSigUDF()
    sync_offset = 2

    ds = lt_ctx.load(
        "dm",
        files=list(sorted(dm_stack_glob)),
        nav_shape=(4, 2),
        io_backend=io_backend,
    )

    result = lt_ctx.run_udf(dataset=ds, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    ds_with_offset = lt_ctx.load(
        "dm",
        files=list(sorted(dm_stack_glob)),
        nav_shape=(4, 2),
        sync_offset=sync_offset,
        io_backend=io_backend,
    )

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.shape.nav.size - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_negative_sync_offset(lt_ctx, io_backend, dm_stack_glob):
    udf = SumSigUDF()
    sync_offset = -2

    ds = lt_ctx.load(
        "dm",
        files=list(sorted(dm_stack_glob)),
        nav_shape=(4, 2),
        io_backend=io_backend,
    )

    result = lt_ctx.run_udf(dataset=ds, udf=udf)
    result = result['intensity'].raw_data[:ds._meta.shape.nav.size - abs(sync_offset)]

    ds_with_offset = lt_ctx.load(
        "dm",
        files=list(sorted(dm_stack_glob)),
        nav_shape=(4, 2),
        sync_offset=sync_offset,
        io_backend=io_backend,
    )

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_missing_frames(lt_ctx, io_backend, dm_stack_glob):
    """
    there can be some frames missing at the end
    """
    # one full row of additional frames in the data set than the number of files
    nav_shape = (3, 5)

    ds = lt_ctx.load(
        "dm",
        files=list(sorted(dm_stack_glob)),
        nav_shape=nav_shape,
        io_backend=io_backend,
    )

    ds.set_num_cores(4)

    tileshape = Shape(
        (1,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p._start_frame == 11
    assert p._num_frames == 4
    assert p.slice.origin == (11, 0, 0)
    assert p.slice.shape[0] == 4
    assert t.tile_slice.origin == (9, 0, 0)
    assert t.tile_slice.shape[0] == 1


def test_scheme_too_large(default_dm):
    partitions = default_dm.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_dm.shape.sig),
        sig_dims=default_dm.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_dm.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_dm.shape.sig)
    for _ in tiles:
        pass


def test_offset_smaller_than_image_count(lt_ctx, dm_stack_glob):
    sync_offset = -12

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "dm",
            files=list(sorted(dm_stack_glob)),
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-10, 10\), which is \(-image_count, image_count\)"
    )


def test_offset_greater_than_image_count(lt_ctx, dm_stack_glob):
    sync_offset = 12

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "dm",
            files=list(sorted(dm_stack_glob)),
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-10, 10\), which is \(-image_count, image_count\)"
    )


@pytest.mark.slow
def test_reshape_nav(lt_ctx, dm_stack_glob):
    udf = SumSigUDF()
    files = list(sorted(dm_stack_glob))

    ds_with_1d_nav = lt_ctx.load("dm", files=files, nav_shape=(8,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("dm", files=files, nav_shape=(4, 2))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("dm", files=files, nav_shape=(2, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


def test_incorrect_sig_shape(lt_ctx, dm_stack_glob):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "dm",
            files=list(sorted(dm_stack_glob)),
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 14238980"
    )


def test_scan_size_deprecation(lt_ctx, dm_stack_glob):
    scan_size = (2, 2)

    with pytest.warns(FutureWarning):
        ds = lt_ctx.load(
            "dm",
            files=list(sorted(dm_stack_glob)),
            scan_size=scan_size,
        )
    assert tuple(ds.shape) == (2, 2, 3838, 3710)


def test_compare_backends(lt_ctx, default_dm, buffered_dm):
    x = random.choice(range(default_dm.shape.nav[0]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_dm,
        x=x,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_dm,
        x=x,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_dm, direct_dm):
    x = random.choice(range(default_dm.shape.nav[0]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_dm,
        x=x,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_dm,
        x=x,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_dm, buffered_dm, as_sparse):
    roi = np.zeros(default_dm.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_dm, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_dm, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


def test_nonstack_files(lt_ctx):
    files = '/path/to/dm/file.dm3'

    with pytest.raises(DataSetException):
        lt_ctx.load(
            "dm",
            files=files,
        )


def test_stack_in_an_inline_pickle(default_dm, lt_ctx):
    # make sure the __new__ magic works with pickle
    import cloudpickle
    ds = cloudpickle.loads(cloudpickle.dumps(default_dm))
    lt_ctx.run_udf(dataset=ds, udf=SumUDF())


def test_load_stack_dd(local_cluster_ctx, dm_stack_glob):
    files = dm_stack_glob
    ds = local_cluster_ctx.load("dm", files=files, same_offset=True)
    ds.check_valid()


def test_bad_params(ds_params_tester, standard_bad_ds_params, dm_3d_glob):
    for params in standard_bad_ds_params:
        ds_params_tester("dm", files=dm_3d_glob, **params)
